import ij.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import ij.gui.*;
import ij.measure.*;
import ij.plugin.MeasurementsWriter;
import java.util.*;

//Particles8 by Gabriel Landini, G.Landini@bham.ac.uk
//1.0 never released
//1.1 background is white when labelling black particles, cleaned code
//1.2 added centre of mass & coordinated. Cleaned mess with newcolor and 253
//1.3 fixed string.equals instead of ==
//1.4 fixed problem with wrong labelling 253 if deleting borders
//1.5 fixed results table erased when no output was selected
//1.6 changed dialog entries similar to those in Particle Analyzer:
// "min" to "Minimum"
// "max" to "Maximum"
// "display" to "Show"
// "Output Particle stats" to "Display Results"
// "Ignore" to "Exclude"
// "Perimeter" to "Perim."
//  data is sent to ResultsTable so it can be read by macros
//  asks before overwriting Results
//  added option to overwrite Results (so macros do not stop)
// 1.7 added Feret's diameter, coordinates and angle with horizontal
// 1.8 30/Apr/2005 fixed a bug to properly delete single pixels at borders if "exclude edge particles" is checked
// 1.9 changed some names to avoid problems with SPSS import
// 1.10 2/Dec/2005 test for squared pixels from calibration values

public class Particles4_ implements PlugInFilter {
	protected String selectedOption;
	protected boolean doIwhite;
	protected boolean doIdeleteborders;
	protected boolean doIminmax;
	protected boolean doIshowstats;
	protected boolean doIlabel;
	protected int stk=0, mi,ma, totp=-1;
	protected ResultsTable rt= ResultsTable.getResultsTable();

	public int setup(String arg, ImagePlus imp) {

		//ImagePlus imp = WindowManager.getCurrentImage();
		if (imp==null)
			return DONE;
		Calibration cal = imp.getCalibration();
		if(cal.pixelWidth!=cal.pixelHeight){
			IJ.error("Image pixels are not squared.");
			return DONE;
		}
		ImageStatistics stats;
		stats=imp.getStatistics();
		if (stats.histogram[0]+stats.histogram[255]!=stats.pixelCount){
			IJ.error("8-bit binary image ( 0 and 255) required.");
			return DONE;
		}

		String header="Slice\tNumber\tXStart\tYStart\tPerim\tArea\tPixels\tXM\tYM\tFeret\tFeretX1\tFeretY1\tFeretX2\tFeretY2\tFAngle";
		if (arg.equals("about"))
			{showAbout(); return DONE;}
		GenericDialog gd = new GenericDialog("Particles4", IJ.getInstance());
		gd.addMessage("Particles4 v1.10 by G. Landini");
		gd.addMessage("This plugin deals ony with \'square pixel\' images!");
		gd.addCheckbox("White particles on black background",true);
		gd.addCheckbox("Exclude edge particles",false);
		gd.addCheckbox("Label particles",false);
		String [] DisplayOption={"Particles", "Coordinates", "CentreOfMass"};
		gd.addChoice("Show", DisplayOption, DisplayOption[0]);
		gd.addCheckbox("Filter by size (pixels)",false);
		gd.addNumericField ("Minimum size (Pixels):",  0, 0);
		gd.addNumericField ("Maximum size (Pixels):",  999999, 0);
		gd.addCheckbox("Display results",true);
		gd.addCheckbox("Overwrite results",false);

		gd.showDialog();
		if (gd.wasCanceled())
			return DONE;

		doIwhite = gd.getNextBoolean ();
		doIdeleteborders = gd.getNextBoolean ();
		doIlabel = gd.getNextBoolean ();
		selectedOption = gd.getNextChoice ();
		doIminmax = gd.getNextBoolean ();
		mi = (int)gd.getNextNumber();
		ma = (int)gd.getNextNumber();
		doIshowstats = gd.getNextBoolean ();

		if ( !gd.getNextBoolean()){
			if (rt.getCounter()>0) {
				SaveChangesDialog di = new SaveChangesDialog(IJ.getInstance(), "Save "+rt.getCounter()+" measurements?");
				if (di.cancelPressed())
					return DONE;
				else if (di.savePressed())
					new MeasurementsWriter().run("");
			}
		}

		if (doIshowstats)
			IJ.setColumnHeadings(header);
		IJ.run("Clear Results");
		return DOES_8G+DOES_STACKS;
	}

	public void run(ImageProcessor ip) {
		stk++;
		int xe = ip.getWidth();
		int ye = ip.getHeight();
		int X = xe-1, Y = ye-1;
		int [] xd = {1,0,-1,0};
		int [] yd = {0,1,0,-1};
		int [] g = {3,0,1,2};

		int x=0, y=0, x1, y1, q, uy, dy, rx, lx, pixs, fig, x2, y2, py, px, nn, newcol, m10, m01, f1, f2, fx1=-1, fx2=-1, fy1=-1, fy2=-1;
		boolean b;
		double nz, narea, mcx, mcy, feret, tferet, rads=180/Math.PI, angle=0.0;
		//ResultsTable rt= ResultsTable.getResultsTable();
		Vector vx = new Vector();
		Vector vy = new Vector();


		if (doIwhite==false){
			for(y=0;y<ye;y++) {
				for(x=0;x<xe;x++)
					ip.putPixel(x,y,255-ip.getPixel(x,y));
			}
		}

	 	fig=0;
		// find first scanned pixel
		for (y=0; y<ye; y++){
			for (x=0; x<xe; x++){
				if (ip.getPixel(x,y)==255){
					
					vx.clear();
					vy.clear();
					
					vx.add(new Integer(x));
					vy.add(new Integer(y));
					
					x1=x;
					y1=y;
					//bounding box
					uy=y;//upper y
					dy=0;//lower y
					rx=0;//right x
					lx=xe;//left x

					nz=0;
					narea=0;
					pixs=0;

					//single pixel?
					if(ip.getPixel(x+1,y)+ip.getPixel(x,y+1)==0){
						if ((doIminmax && (mi>0))|| (doIdeleteborders && (x==0 || x==X || y==0 || y==Y))){
							ip.putPixel(x,y,0);
						}
						else{
							ip.putPixel(x,y,((fig  % 251) + 1)); //new colour if labelling
							pixs=1;
							fig++;
							totp++;
							if (doIshowstats) IJ.write(stk + "\t"+fig+"\t" + x + "\t"+ y +"\t"+0+"\t"+0+"\t"+pixs+"\t"+x+"\t"+y+"\t"+0+"\t"+x+"\t"+y+"\t"+x+"\t"+y+"\t"+0);
							rt.incrementCounter();
							rt.setValue("Slice", totp, stk);
							rt.setValue("Number", totp, fig);
							rt.setValue("XStart", totp, x);
							rt.setValue("YStart", totp, y);
							rt.setValue("Perim", totp, 0);
							rt.setValue("Area", totp, 0);
							rt.setValue("Pixels", totp, pixs);
							rt.setValue("XM", totp, x);
							rt.setValue("YM", totp, y);
							rt.setValue("Feret", totp, 0);
							rt.setValue("FeretX1", totp, x);
							rt.setValue("FeretY1", totp, y);
							rt.setValue("FeretX2", totp, x);
							rt.setValue("FeretY2", totp, y);
							rt.setValue("FAngle", totp, 0);	
						}
					}
					else{ //go around the edge
						x1=x;
						y1=y;
						ip.putPixel(x,y,254);
						q=0;
						while(true){
							if (ip.getPixel(x1+xd[q],y1+yd[q])>0){
								x2=x1;
								y2=y1;
								x1+=xd[q];
								y1+=yd[q];
								narea+=(double)((y1+y2)*(x2-x1))*.5;
								ip.putPixel(x1,y1,254);
								//update coordinates of bounding box
								if(x1 > rx) rx = x1;
								if(x1 < lx) lx = x1;
								if(y1 > dy) dy = y1;
								if(y1 < uy) uy = y1;
								nz++; //perimeter
								q=g[q];

								//check for *that* unique pixel configuration in 4 conn
								if (x1==x){
									if (y1==y){
										if (ip.getPixel(x,y+1)!=255){
											break;
										}
									}
								}
								vx.add(new Integer(x1));
								vy.add(new Integer(y1));
							}
							else
								q=(q+1)%4;
						}
						
						// feret diameter here
						feret=0.0;
						tferet=0.0;
						for (f1=0;f1<vx.size();f1++){
							for (f2=f1;f2<vx.size();f2++){
								tferet=	Math.pow(((Integer)(vx.get(f1))).intValue() - ((Integer)(vx.get(f2))).intValue(),2) + Math.pow(((Integer)(vy.get(f1))).intValue() - ((Integer)(vy.get(f2))).intValue(),2);
								if (tferet>feret){
									feret=tferet;
									fx1=((Integer)(vx.get(f1))).intValue();
									fy1=((Integer)(vy.get(f1))).intValue();
									fx2=((Integer)(vx.get(f2))).intValue();
									fy2=((Integer)(vy.get(f2))).intValue();
								}
							}
						}
						feret=Math.sqrt(feret);
						//IF f2x - fx = 0 THEN THETA = 90 ELSE THETA = -INT(ATN((f2y - fy) / (f2x - fx)) * 57.29583)
						if (fx2==fx1)
							angle=90.0;
						else
							angle=-Math.atan((double)(fy2-fy1)/(double)(fx2-fx1))*rads;

						//label from top to bottom and vice versa until no new
						//pixels are labelled. No need to check all 8 neighbours in
						//each pass, but only the previous 4 in the scanning direction
						b=true;
						while(b){
							b=false;
							//scan bounding box from top left to bottom right
							for(py=uy+1;py<dy;py++) {
								for(px=lx+1;px<rx;px++) {
									if (ip.getPixel(px,py)==255){
										if(ip.getPixel(px,py-1)==254 ||
										  ip.getPixel(px-1,py)==254) {
											ip.putPixel(px,py,254);
											b=true; // this pixel is connected to the border
										}
									}
								}
							}
							//scan bounding box bottom right to top left
							for(py=dy-1;py>=uy;py--) {
								for(px=rx-1;px>=lx;px--) {
									if (ip.getPixel(px,py)==255){
										if(ip.getPixel(px+1,py)==254 ||
										   ip.getPixel(px,py+1)==254) {
											ip.putPixel(px,py,254);
											b=true; // this pixel is connected to the border
										}
									}
								}
							}
						}
						newcol = (fig  % 251) + 1; //new colour if labelling
						//a2 = newcol + 1;
						m10 = 0;
						m01 = 0;
						mcx=0;
						mcy=0;
						if(doIminmax || doIdeleteborders)
							nn = 253; //must have an extra colour not used elsewhere
						 else
						 	nn = newcol;
						for(py=uy;py<=dy;py++){
							for(px=lx;px<=rx;px++){
								if(ip.getPixel(px,py)== 254){
									pixs+=1;
									m10+=px;
									m01+=py;
									ip.putPixel(px,py,nn);
								}
							}
						}
						mcx = ((double)m10 / pixs);
						mcy = ((double)m01 / pixs);
						//momentum
						//System.out.println("Debug:"+xe+" "+ye+":   "+ uy+ " "+lx+" "+dy+" "+rx);
						// delete border touching particles
						if(doIdeleteborders && (uy==0 || lx==0 || dy==ye-1 || rx==xe-1)){
							for(py=uy;py<=dy;py++){
								for(px=lx;px<=rx;px++){
									if(ip.getPixel(px,py)== nn)
										ip.putPixel(px,py,0);
								}
							}
						}
						else{
							if (doIminmax || doIdeleteborders){
								if (pixs<mi || pixs>ma){ // erase (filter) particles
									for(py=uy;py<=dy;py++){
										for(px=lx;px<=rx;px++){
											if(ip.getPixel(px,py)== nn)
												ip.putPixel(px,py,0);
										}
									}
								}
								else{ //particle not to be deleted...

									if (selectedOption.equals("Coordinates") || selectedOption.equals("CentreOfMass")){ // ...but put centre of mass
										for(py=uy;py<=dy;py++){
											for(px=lx;px<=rx;px++){
												if(ip.getPixel(px,py)== nn)
													ip.putPixel(px,py,0); //clear the particle...
											}
										}
										if(selectedOption.equals("CentreOfMass"))
											ip.putPixel((int)Math.floor(mcx+0.5),(int)Math.floor(mcy+0.5),newcol);//... and put centre of mass
										if(selectedOption.equals("Coordinates"))
											ip.putPixel(x,y,newcol);//... and put start coordinates
									}
									else{ //change colour from 253 to newcol
										for(py=uy;py<=dy;py++){
											for(px=lx;px<=rx;px++){
												if(ip.getPixel(px,py)== nn) //253
													ip.putPixel(px,py,newcol);
											}
										}
									}
									fig++;
									totp++;
									if (doIshowstats) IJ.write(stk + "\t"+fig+"\t" + x + "\t"+ y +"\t"+(float)nz+"\t"+(float)narea+"\t"+pixs+"\t"+(float)mcx+"\t"+(float)mcy+"\t"+(float)feret+"\t"+fx1+"\t"+fy1+"\t"+fx2+"\t"+fy2+"\t"+(float)angle);

									rt.incrementCounter();
									rt.setValue("Slice", totp , stk);
									rt.setValue("Number", totp, fig);
									rt.setValue("XStart", totp, x);
									rt.setValue("YStart", totp, y);
									rt.setValue("Perim", totp, (float)nz);
									rt.setValue("Area", totp, (float)narea);
									rt.setValue("Pixels", totp, pixs);
									rt.setValue("XM", totp, (float)mcx);
									rt.setValue("YM", totp, (float)mcy);
									rt.setValue("Feret", totp, (float)feret);
									rt.setValue("FeretX1", totp, fx1);
									rt.setValue("FeretY1", totp, fy1);
									rt.setValue("FeretX2", totp, fx2);
									rt.setValue("FeretY2", totp, fy2);
									rt.setValue("FAngle", totp, (float)angle);	
								}
							}
							else{ //particle is already the correct colour
								if (selectedOption.equals("Coordinates") || selectedOption.equals("CentreOfMass")){
									for(py=uy;py<=dy;py++){
										for(px=lx;px<=rx;px++){
											if(ip.getPixel(px,py)== newcol) //253 erase...
												ip.putPixel(px,py,0);
										}
									}
									if(selectedOption.equals("CentreOfMass"))
										ip.putPixel((int)Math.floor(mcx+0.5),(int)Math.floor(mcy+0.5),nn);//... and put centre of mass
									if(selectedOption.equals("Coordinates"))
										ip.putPixel(x,y,nn);//... and put start coordinates
								}
								fig++;
								totp++;
								if (doIshowstats) IJ.write(stk + "\t"+fig+"\t" + x + "\t"+ y +"\t"+(float)nz+"\t"+(float)narea+"\t"+pixs+"\t"+(float)mcx+"\t"+(float)mcy+"\t"+(float)feret+"\t"+fx1+"\t"+fy1+"\t"+fx2+"\t"+fy2+"\t"+(float)angle);

								rt.incrementCounter();
								rt.setValue("Slice", totp, stk);
								rt.setValue("Number", totp, fig);
								rt.setValue("XStart", totp, x);
								rt.setValue("YStart", totp, y);
								rt.setValue("Perim", totp, (float)nz);
								rt.setValue("Area", totp, (float)narea);
								rt.setValue("Pixels", totp, pixs);
								rt.setValue("XM", totp, (float)mcx);
								rt.setValue("YM", totp, (float)mcy);
								rt.setValue("Feret", totp, (float)feret);
								rt.setValue("FeretX1", totp, fx1);
								rt.setValue("FeretY1", totp, fy1);
								rt.setValue("FeretX2", totp, fx2);
								rt.setValue("FeretY2", totp, fy2);
								rt.setValue("FAngle", totp, (float)angle);	
							}
						}
					}
				}
			}
		}

		//if not labelling, change the colour back to 255
		if (doIlabel==false){
			for(y=0;y<ye;y++) {
				for(x=0;x<xe;x++){
					if (ip.getPixel(x,y)>0)
						ip.putPixel(x,y,255);
				}
			}
			// if white particles, invert
			if (doIwhite==false){
				for(y=0;y<ye;y++) {
					for(x=0;x<xe;x++)
						ip.putPixel(x,y,255-ip.getPixel(x,y));
				}
			}
		}
		else{
			if (doIwhite==false){
				for(y=0;y<ye;y++) {
					for(x=0;x<xe;x++){
						if (ip.getPixel(x,y)==0)
							ip.putPixel(x,y,255); //white background
					}
				}
			}
		}
	}



	void showAbout() {
		IJ.showMessage("About Particles4_...",
		"Particles4 by Gabriel Landini,  G.Landini@bham.ac.uk\n"+
		"ImageJ plugin for estimating various statistics of 4-connected\n"+
		"particles.\n"+
		"WARNING!: This plugin does not return the same values as the original\n"+
		"ImageJ \'Analyze Particles\' command because it uses a different concept.\n"+
		"Here, \'Perimeter\' and \'Area\' are measured based on the polygon given by\n"+
		"the centres of the boundary pixels. \'Area\' disregards \'holes\' in the particles\n"+
		"(i.e. it returns the area inside the boundary), but \'Pixels\' does not.\n"+
		"Note that 1 pixel particles have 0 area, etc.\n"+
		"Filtered particles are deleted from the image");

	}

}
